/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File handler.c */

/* Thus speaketh ye mighty CVS:
 * $Id: handler.c,v 1.2 1999/05/02 23:21:36 david Exp $ */

/* This program contains the main function of the handler threads. */

/* CVS identity of this file. */
static char id[] = "$Id: handler.c,v 1.2 1999/05/02 23:21:36 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */

#include <unistd.h>		/* UNIX standard functions */

#include <pthread.h>		/* Posix threads */


#include "server.h"

#include "lgdproto.h"

/* Main functions */

void
reply_error (struct lsock_s *sck, unsigned long callback,
	     unsigned long errcode)
     /* Return an error code. */
{
  char data[20];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_ERR);
  *(unsigned long *)(data+4) = htonl (20);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  *(unsigned long *)(data+16) = htonl (errcode);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
reply_simple (struct lsock_s *sck, unsigned long callback,
	      unsigned long reply)
     /* Return just a 32-bit quantity. */
{
  char data[20];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_SYN);
  *(unsigned long *)(data+4) = htonl (20);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  *(unsigned long *)(data+16) = htonl (reply);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
do_greet (struct lsock_s *sck)
{
  char data[24];

  sck->state = CONN_STATE_NORMAL;
  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_CON);
  *(unsigned long *)(data+4) = htonl (24);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = htonl (LGD_PROTO_MAGIC);
  *(unsigned long *)(data+16) = htonl (LGD_PROTO_VERNUM);
  *(unsigned long *)(data+20) = htonl (1);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
do_request (struct lsock_s *sck)
     /* Handle a request on the given connection. */
{
  unsigned long len,req,callback;
#if ( DEBUG_LEVEL >= 3 )
  fprintf (stderr, "[%0#10lx]Considering request.\n",
	   pthread_self ());
#endif /* DEBUG_LEVEL >= 3 */
  if ( sck->buflen < 8 )
    /* We don't have the request code and length yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	sck->state = CONN_STATE_DEAD;
      sck->handwork = 0;
      return;
    }
  req = ntohl (*(unsigned long *)(sck->buffer));
  len = ntohl (*(unsigned long *)(sck->buffer+4));
  if ( sck->buflen >= 16 )
    callback = *(unsigned long *)(sck->buffer+12);
  else
    callback = htonl (0); /* Yes, this is htonl() and not ntohl(). */
  if ( len == 0 )
    /* XXX - What should we do in this case?  I don't know.  Let's
     * just silently close the connection.  Note that we allow request
     * lengths less than 16, which have pretty weird effects... */
    {
      if ( sck->sopen )
	shutdown (sck->s, 0);
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 0;
      return;
    }
  if ( len > BUFSIZE )
    /* Request is too long to fit in buffer no matter what.
     * Complain. */
    {
#if ( DEBUG_LEVEL >= 1 )
      fprintf (stderr, "[%0#10lx]Request is too long.\n",
	       pthread_self ());
#endif /* DEBUG_LEVEL >= 1 */
      reply_error (sck, callback, LGD_PROTO_ERR_TOOLONG);
      if ( sck->sopen )
       	shutdown (sck->s, 0);
      /* Someone else will take care of closing s. */
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 0;
      /* XXX - Think of something smarter. */
      return;
    }
  if ( sck->buflen < len )
    /* We don't have the full request yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	{
	  len = sck->buflen;
	  /* Redo with what we have. */
	  return;
	}
      sck->handwork = 0;
      return;
    }
#if ( DEBUG_LEVEL >= 2 )
  fprintf (stderr, "[%0#10lx]Now handling request: req=%0#10lx, len=%0#10lx\n",
	   pthread_self (), req, len);
#endif /* DEBUG_LEVEL >= 2 */
  switch ( req )
    {
    case LGD_PROTO_REQ_NOP:
      /* NOP */
      break;
    case LGD_PROTO_REQ_ACK:
      /* ACK? */
      reply_simple (sck, callback, 1);
      break;
    case LGD_PROTO_REQ_MAGIC:
      /* MAGIC */
      reply_simple (sck, callback, LGD_PROTO_MAGIC);
      break;
    case LGD_PROTO_REQ_VERNUM:
      /* VERNUM */
      reply_simple (sck, callback, LGD_PROTO_VERNUM);
      break;
    case LGD_PROTO_REQ_DIE:
      /* DIE */
      exit (1);
      break;
    default:
      reply_error (sck, callback, LGD_PROTO_ERR_UNKNOWN);
      break;
    }
  memmove (sck->buffer, sck->buffer+len, len);
  sck->buflen -= len;
  /* *Don't* set sck->handwork back to 0, because there may be another
   * request in the buffer. */
#if ( DEBUG_LEVEL >= 2 )
  fprintf (stderr, "[%0#10lx]Request handled.\n",
	   pthread_self ());
#endif /* DEBUG_LEVEL >= 2 */
}

void
handler_loop (void)
     /* This function performs a loop that handles the requests that
      * the listener threads have put in the connection buffers.  See
      * NOTE0008 for an overview of how this function works (wrt
      * sleeping, that is). */
{
  int i;
  struct lsock_s *sck;
  char some_handwork;
  long prior_count;

  safe_pthread_mutex_lock (&handler_wake_mutex);
  i = 0;
  some_handwork = 0;
  prior_count = handler_wake_count;
  safe_pthread_mutex_unlock (&handler_wake_mutex);
  while (1)
    {
      /* The trick here is that we don't want to hold the lsock_mutex
       * for too long as it is a precious resource.  So we do a loop
       * that keeps locking it and unlocking it.  I wonder if this is
       * really better, thought.  Maybe a rwlock would be more
       * appropriate than a mutex (see NOTE0001). */
      safe_pthread_mutex_lock (&lsock_mutex);
      if ( i >= lsock_nb )
	/* Looped around all connections (actually, we may have missed
	 * some since we weren't holdong the lsock_mutex all the time,
	 * but that doesn't really matter, we'll get them next
	 * time). */
	{
	  safe_pthread_mutex_unlock (&lsock_mutex);
	  if ( ! some_handwork )
	    /* Hmmm...  We looped around all connections and found
	     * nothing new at all.  Perhaps we finished our task.  We
	     * need to make sure, though, because we may have missed
	     * some connections.  So we check that nothing has
	     * happened behind our back during this loop. */
	    {
	      safe_pthread_mutex_lock (&handler_wake_mutex);
	      if ( handler_wake_count == prior_count )
		/* Indeed, no new data has arrived during the entire
		 * loop.  So we are quite finished, and we can go back
		 * to sleep.  We will release the handler_wake_mutex
		 * atomically as we start sleeping. */
		goto handler_sleep;
	      /* Keep holding on that mutex, baby. */
	    }
	  else
	    safe_pthread_mutex_lock (&handler_wake_mutex);
	  i = 0;
	  some_handwork = 0;
	  prior_count = handler_wake_count;
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	  sched_yield();
	  continue;
	}
      sck = lsock_tab[i];
      if ( sck->handwork )
	/* [NOTE0007] We perform this test without holding the d_mutex
	 * (still we hold the lsock_mutex otherwise the sck
	 * dereference would be forbidden, see NOTE0001).  This is
	 * because we can't guarantee that we will be able to hold the
	 * d_mutex at all.  Still, we need to have some idea of
	 * whether there actually is something new here.  Perhaps we
	 * will read 1 when in fact it was set to 0: that is
	 * unimportant.  On the other hand, when sck->handwork is set
	 * to 1 by a listener thread, the latter also increases the
	 * handler_wake_count, so if we miss that we will notice
	 * something when we check handler_wake_count.  To summarize,
	 * any value of sck->handwork later than the last time we
	 * unlocked the handler_wake_mutex will do, and we are
	 * guaranteed to get such a value.  (Moreover, we can't have
	 * any byte tearing because sck->handwork is a char.)  So this
	 * is OK.  (Would somebody please confirm...) */
	some_handwork = 1;
      if ( safe_pthread_mutex_trylock (&sck->d_mutex) )
	{
	  /* Someone is doing something with this connection already.
	   * We go on. */
	  safe_pthread_mutex_unlock (&lsock_mutex);
	  i++;
	  sched_yield();
	  continue;
	}
      /* Now we have the d_mutex so we can release the lsock_mutex to
       * access sck (see NOTE0001). */
      safe_pthread_mutex_unlock (&lsock_mutex);
      /* Do everything */
      switch ( sck->state )
	{
	case CONN_STATE_NEW:
	  do_greet (sck);
	  break;
	case CONN_STATE_NORMAL:
	  do_request (sck);
	  break;
	case CONN_STATE_DEAD:
	  /* Dead connections are ignored. */
	  sck->handwork = 0;
	  /* XXX - Actually, *can* this happen?  Isn't sck->handwork
	   * necessarily 0 for a dead connection? */
	  break;
	}
      safe_pthread_mutex_unlock (&sck->d_mutex);
      i++;
      continue;
    handler_sleep:
      /* When we arrive here we *must* be owning the
       * handler_wake_mutex. */
      pthread_cond_wait (&handler_wake_cond, &handler_wake_mutex);
      safe_pthread_mutex_unlock (&handler_wake_mutex);
    }
}
